# skedge

[![Crates.io](https://img.shields.io/crates/v/skedge.svg)](https://crates.io/crates/skedge)
[![rust action](https://github.com/deciduously/skedge/actions/workflows/rust.yml/badge.svg)](https://github.com/deciduously/skedge/actions/workflows/rust.yml)
[![docs.rs](https://img.shields.io/docsrs/skedge/0.1.0)](https://docs.rs/skedge/0.1.0)

Rust single-process scheduling.  Ported from [`schedule`](https://github.com/dbader/schedule) for Python, in turn inspired by [`clockwork`](https://github.com/Rykian/clockwork) (Ruby), and ["Rethinking Cron"](https://adam.herokuapp.com/past/2010/4/13/rethinking_cron/) by [Adam Wiggins](https://github.com/adamwiggins).

## Usage

Documentation can be found on [docs.rs](https://docs.rs/skedge).

This library uses the Builder pattern to define jobs.  Instantiate a fresh `Scheduler`, then use the `every()` and `every_single()` functions to begin defining a job.  Finalize configuration by calling `Job::run()` to add the new job to the scheduler.  The `Scheduler::run_pending()` method is used to fire any jobs that have arrived at their next scheduled run time.  Currently, precision can only be specified to the second, no smaller.

```rust
use chrono::Local;
use skedge::{every, Scheduler};
use std::thread::sleep;
use std::time::Duration;

fn greet(name: &str) {
    println!("Hello {}, it's {}!", name, Local::now().to_rfc2822());
}

fn main() -> Result<(), Box<dyn std::error::Error>> {
    let mut schedule = Scheduler::new();

    every(2)
        .to(8)?
        .seconds()?
        .until(Local::now() + chrono::Duration::seconds(30))?
        .run_one_arg(&mut schedule, greet, "Good-Looking")?;

    println!("Starting at {}", Local::now());
    loop {
        if let Err(e) = schedule.run_pending() {
            eprintln!("Error: {}", e);
        }
        sleep(Duration::from_secs(1));
    }
}
```

Check out the [example script](https://github.com/deciduously/skedge/blob/main/examples/basic.rs) to see more configuration options.  Try `cargo run --example readme` or `cargo run --example basic` to see it in action.

## Development

Clone this repo.  See [`CONTRIBUTING.md`](https://github.com/deciduously/skedge/blob/main/CONTRIBUTING.md) for contribution guidelines.

### Dependencies
 
* **Stable [Rust](https://www.rust-lang.org/tools/install)**:  The default stable toolchain is fine.  Obtainable via `rustup` using the instructions at this link.

### Crates

* [chrono](https://github.com/chronotope/chrono) - Date and time handling
* [log](https://github.com/rust-lang/log) - Logging
* [pretty_env_logger](https://github.com/seanmonstar/pretty-env-logger) - Pretty logging
* [lazy_static](https://github.com/rust-lang-nursery/lazy-static.rs) - Lazily evaluated statics
* [rand](https://rust-random.github.io/book/) - Random number generation
* [regex](https://github.com/rust-lang/regex) - Regular expressions
* [thiserror](https://github.com/dtolnay/thiserror) - Error derive macro

#### Development-Only

 * [pretty_assertions](https://github.com/colin-kiegel/rust-pretty-assertions) - Colorful assertion output